 /* Excelsior theme engine
 * Copyright (C) 2005 Richard Stellingwerff
 * Copyright (C) 2007 Benjamin Berg
 * Copyright (C) 2007 Andrea Cimitan
 * Copyright (C) 2008 Christopher Bratusek
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

#include <gtk/gtk.h>
#include <cairo.h>
#include <math.h>
#include <string.h>

#include "ge-support.h"
#include "excelsior_style.h"
#include "excelsior_rc_style.h"
#include "excelsior_draw.h"
#include "support.h"

#include "../config.h"

/* #define DEBUG 1 */

#define DETAIL(xx)   ((detail) && (!strcmp(xx, detail)))
#define CHECK_HINT(xx) (ge_check_hint ((xx), EXCELSIOR_RC_STYLE ((style)->rc_style)->hint, widget))

#define DRAW_ARGS    GtkStyle       *style, \
                     GdkWindow      *window, \
                     GtkStateType    state_type, \
                     GtkShadowType   shadow_type, \
                     GdkRectangle   *area, \
                     GtkWidget      *widget, \
                     const gchar    *detail, \
                     gint            x, \
                     gint            y, \
                     gint            width, \
                     gint            height

#ifdef HAVE_ANIMATION
#include "animation.h"
#endif

#define STYLE_FUNCTION(function) (excelsior_style_class->style_functions[EXCELSIOR_STYLE (style)->style].function)

static excelsiorStyleClass *excelsior_style_class;
static GtkStyleClass *excelsior_parent_class;

static void
excelsior_set_widget_parameters (const GtkWidget      *widget,
                                  const GtkStyle       *style,
                                  GtkStateType          state_type,
                                  WidgetParameters     *params)
{
	params->style_functions = &(excelsior_style_class->style_functions[EXCELSIOR_STYLE (style)->style]);

	if (widget && GE_IS_ENTRY (widget))
		state_type = GTK_WIDGET_STATE (widget);

	params->active        = (state_type == GTK_STATE_ACTIVE);
	params->prelight      = (state_type == GTK_STATE_PRELIGHT);
	params->disabled      = (state_type == GTK_STATE_INSENSITIVE);
	params->state_type    = (excelsiorStateType)state_type;
	params->corners       = CR_CORNER_ALL;
	params->ltr           = ge_widget_is_ltr ((GtkWidget*)widget);
	params->focus         = widget && GTK_WIDGET_HAS_FOCUS (widget);
	params->is_default    = widget && GE_WIDGET_HAS_DEFAULT (widget);
	params->enable_shadow = FALSE;
	params->radius        = EXCELSIOR_STYLE (style)->radius;
	params->random_effect = EXCELSIOR_STYLE (style)->random_effect;
	params->random_opacity = EXCELSIOR_STYLE (style)->random_opacity;
	params->mw_gradient   = EXCELSIOR_STYLE (style)->mw_gradient;
	params->aqua_tabs     = EXCELSIOR_STYLE (style)->aqua_tabs;
	params->pellucid_scales = EXCELSIOR_STYLE (style)->pellucid_scales;
	params->scale_trough_size = EXCELSIOR_STYLE (style)->scale_trough_size;
	params->tooltip_style = EXCELSIOR_STYLE (style)->tooltip_style;
	params->xthickness    = style->xthickness;
	params->ythickness    = style->ythickness;
	params->parentbg      = EXCELSIOR_STYLE (style)->colors.bg[state_type];
	excelsior_get_parent_bg (widget, &params->parentbg);
	params->grip_opacity  = EXCELSIOR_STYLE (style)->grip_opacity;
	params->grip_style    = EXCELSIOR_STYLE (style)->grip_style;
	params->trough_visible = EXCELSIOR_STYLE (style)->trough_visible;
}

static void
excelsior_style_draw_flat_box (DRAW_ARGS)
{
	if (detail &&
	    state_type == GTK_STATE_SELECTED && (
	    !strncmp ("cell_even", detail, 9) ||
	    !strncmp ("cell_odd", detail, 8)))
	{
		WidgetParameters params;
		excelsiorStyle  *excelsior_style;
		excelsiorColors *colors;
		cairo_t          *cr;

		CHECK_ARGS
		SANITIZE_SIZE

		excelsior_style = EXCELSIOR_STYLE (style);
		excelsior_set_widget_parameters (widget, style, state_type, &params);
		colors = &excelsior_style->colors;
		cr = ge_gdk_drawable_to_cairo (window, area);

		STYLE_FUNCTION (draw_selected_cell) (cr, colors, &params, x, y, width, height);

		cairo_destroy (cr);
	}
	else if (DETAIL ("icon_view_item"))
	{
		WidgetParameters params;
		excelsiorStyle  *excelsior_style;
		excelsiorColors *colors;
		cairo_t          *cr;

		CHECK_ARGS
		SANITIZE_SIZE

		excelsior_style = EXCELSIOR_STYLE (style);
		excelsior_set_widget_parameters (widget, style, state_type, &params);
		colors = &excelsior_style->colors;
		cr = ge_gdk_drawable_to_cairo (window, area);

		STYLE_FUNCTION (draw_icon_view_item) (cr, colors, &params, x, y, width, height);

		cairo_destroy (cr);
	}
	else if (DETAIL ("tooltip"))
	{
		WidgetParameters params;
		excelsiorStyle  *excelsior_style;
		excelsiorColors *colors;
		cairo_t          *cr;

		CHECK_ARGS
		SANITIZE_SIZE

		excelsior_style = EXCELSIOR_STYLE (style);
		excelsior_set_widget_parameters (widget, style, state_type, &params);
		colors = &excelsior_style->colors;
		cr = ge_gdk_drawable_to_cairo (window, area);

		STYLE_FUNCTION (draw_tooltip) (cr, colors, &params, x, y, width, height);

		cairo_destroy (cr);
	}
	else if (GE_IS_ENTRY(widget))
	{
		excelsiorStyle  *excelsior_style = EXCELSIOR_STYLE (style);
		excelsiorColors *colors = &excelsior_style->colors;
		cairo_t          *cr     = ge_gdk_drawable_to_cairo (window, area);

		CHECK_ARGS
		SANITIZE_SIZE

		WidgetParameters params;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		STYLE_FUNCTION (draw_entry) (cr, colors, &params, x-2, y-2, width+4, height+4);
	}
	else if (GE_IS_WINDOW(widget))
	{
		excelsiorStyle  *excelsior_style = EXCELSIOR_STYLE (style);
		excelsiorColors *colors = &excelsior_style->colors;
		cairo_t          *cr     = ge_gdk_drawable_to_cairo (window, area);

		CHECK_ARGS
		SANITIZE_SIZE

		WidgetParameters params;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		STYLE_FUNCTION (draw_mw_gradient) (cr, colors, &params, x, y, width, height);
	}
	else
	{
		excelsior_parent_class->draw_flat_box (style, window, state_type,
		                                        shadow_type,
		                                        area, widget, detail,
		                                        x, y, width, height);
	}
}

static void
excelsior_style_draw_shadow (DRAW_ARGS)
{
	excelsiorStyle  *excelsior_style = EXCELSIOR_STYLE (style);
	excelsiorColors *colors = &excelsior_style->colors;
	cairo_t          *cr     = ge_gdk_drawable_to_cairo (window, area);

	CHECK_ARGS
	SANITIZE_SIZE

	if ((DETAIL ("entry") && !CHECK_HINT (GE_HINT_TREEVIEW)) ||
	    (DETAIL ("frame") && CHECK_HINT (GE_HINT_COMBOBOX_ENTRY)))
	{
		WidgetParameters params;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		if (GE_IS_ENTRY(widget))
			params.enable_shadow = TRUE;

		if (!GE_IS_BIN(widget) || !GE_IS_BIN(widget->parent))
		STYLE_FUNCTION (draw_entry) (cr, &excelsior_style->colors, &params,
		                             x, y, width, height);
	}
	else if (DETAIL ("frame") && CHECK_HINT (GE_HINT_STATUSBAR))
	{
		WidgetParameters params;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		gtk_style_apply_default_background (style, window, FALSE, state_type,
		                                    area, x, y, width, height);
		if (shadow_type != GTK_SHADOW_NONE)
			STYLE_FUNCTION (draw_statusbar) (cr, colors, &params,
			                                 x, y, width, height);
	}
	else if (DETAIL ("frame"))
	{
		WidgetParameters params;
		FrameParameters  frame;
		frame.shadow  = shadow_type;
		frame.gap_x   = -1;  /* No gap will be drawn */
		frame.border  = &colors->shade[4];

		excelsior_set_widget_parameters (widget, style, state_type, &params);
		params.corners = CR_CORNER_NONE;

		if (widget && !g_str_equal ("XfcePanelWindow", gtk_widget_get_name (gtk_widget_get_toplevel (widget))))
			STYLE_FUNCTION(draw_frame) (cr, colors, &params, &frame,
			                            x, y, width, height);
	}
	else if (DETAIL ("scrolled_window") || DETAIL ("viewport") || detail == NULL)
	{
		WidgetParameters params;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

	}
	else
	{
		WidgetParameters params;
		FrameParameters frame;

		frame.shadow = shadow_type;
		frame.gap_x  = -1;
		frame.border = &colors->shade[5];
		excelsior_set_widget_parameters (widget, style, state_type, &params);
		params.corners = CR_CORNER_ALL;

		STYLE_FUNCTION(draw_frame) (cr, colors, &params, &frame, x, y, width, height);
	}

	cairo_destroy (cr);
}

static void
excelsior_style_draw_box_gap (DRAW_ARGS,
                               GtkPositionType gap_side,
                               gint            gap_x,
                               gint            gap_width)
{
	excelsiorStyle  *excelsior_style = EXCELSIOR_STYLE (style);
	excelsiorColors *colors = &excelsior_style->colors;
	cairo_t          *cr;

	CHECK_ARGS
	SANITIZE_SIZE

	cr = ge_gdk_drawable_to_cairo (window, area);

	if (DETAIL ("notebook"))
	{
		WidgetParameters params;
		FrameParameters  frame;
		gboolean start, end;

		frame.shadow    = shadow_type;
		frame.gap_side  = gap_side;
		frame.gap_x     = gap_x;
		frame.gap_width = gap_width;
		frame.border    = &colors->shade[5];

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		excelsior_get_notebook_tab_position (widget, &start, &end);

		params.corners = CR_CORNER_ALL;
		switch (gap_side)
		{
			case GTK_POS_TOP:
				if (ge_widget_is_ltr (widget))
				{
					if (start)
						params.corners ^= CR_CORNER_TOPLEFT;
					if (end)
						params.corners ^= CR_CORNER_TOPRIGHT;
				}
				else
				{
					if (start)
						params.corners ^= CR_CORNER_TOPRIGHT;
					if (end)
						params.corners ^= CR_CORNER_TOPLEFT;
				}
			break;
			case GTK_POS_BOTTOM:
				if (ge_widget_is_ltr (widget))
				{
					if (start)
						params.corners ^= CR_CORNER_BOTTOMLEFT;
					if (end)
						params.corners ^= CR_CORNER_BOTTOMRIGHT;
				}
				else
				{
					if (start)
						params.corners ^= CR_CORNER_BOTTOMRIGHT;
					if (end)
						params.corners ^= CR_CORNER_BOTTOMLEFT;
				}
			break;
			case GTK_POS_LEFT:
				if (start)
					params.corners ^= CR_CORNER_TOPLEFT;
				if (end)
					params.corners ^= CR_CORNER_BOTTOMLEFT;
			break;
			case GTK_POS_RIGHT:
				if (start)
					params.corners ^= CR_CORNER_TOPRIGHT;
				if (end)
					params.corners ^= CR_CORNER_BOTTOMRIGHT;
			break;
		}

		/* Fill the background with bg[NORMAL] */
		ge_cairo_rounded_rectangle (cr, x, y, width, height, params.radius, params.corners);
		ge_cairo_set_color (cr, &colors->bg[GTK_STATE_NORMAL]);
		cairo_fill (cr);

		STYLE_FUNCTION(draw_frame) (cr, colors, &params, &frame,
		                            x, y, width, height);
	}
	else
	{
		excelsior_parent_class->draw_box_gap (style, window, state_type, shadow_type,
		                                       area, widget, detail,
		                                       x, y, width, height,
		                                       gap_side, gap_x, gap_width);
	}

	cairo_destroy (cr);
}

static void
excelsior_style_draw_extension (DRAW_ARGS, GtkPositionType gap_side)
{
	excelsiorStyle  *excelsior_style = EXCELSIOR_STYLE (style);
	excelsiorColors *colors = &excelsior_style->colors;
	cairo_t          *cr;

	CHECK_ARGS
	SANITIZE_SIZE

	cr = ge_gdk_drawable_to_cairo (window, area);

	if (DETAIL ("tab"))
	{
		WidgetParameters params;
		TabParameters    tab;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		tab.gap_side = (excelsiorGapSide)gap_side;

		switch (gap_side)
		{
			case XCL_GAP_TOP:
				params.corners = CR_CORNER_BOTTOMLEFT | CR_CORNER_BOTTOMRIGHT;
				break;
			case XCL_GAP_BOTTOM:
				params.corners = CR_CORNER_TOPLEFT | CR_CORNER_TOPRIGHT;
				break;
			case XCL_GAP_LEFT:
				params.corners = CR_CORNER_TOPRIGHT | CR_CORNER_BOTTOMRIGHT;
				break;
			case XCL_GAP_RIGHT:
				params.corners = CR_CORNER_TOPLEFT | CR_CORNER_BOTTOMLEFT;
				break;
		}

		STYLE_FUNCTION(draw_tab) (cr, colors, &params, &tab,
		                          x, y, width, height);
	}
	else
	{
		excelsior_parent_class->draw_extension (style, window, state_type, shadow_type, area,
		                                         widget, detail, x, y, width, height,
		                                         gap_side);
	}

	cairo_destroy (cr);
}

static void
excelsior_style_draw_handle (DRAW_ARGS, GtkOrientation orientation)
{
	excelsiorStyle  *excelsior_style = EXCELSIOR_STYLE (style);
	excelsiorColors *colors = &excelsior_style->colors;
	cairo_t          *cr;

	CHECK_ARGS
	SANITIZE_SIZE

	cr = ge_gdk_drawable_to_cairo (window, area);

	if (DETAIL ("handlebox"))
	{
		WidgetParameters params;
		HandleParameters handle;

		excelsior_set_widget_parameters (widget, style, state_type, &params);
		handle.type = XCL_HANDLE_TOOLBAR;
		handle.horizontal = (orientation == GTK_ORIENTATION_HORIZONTAL);

		STYLE_FUNCTION(draw_handle) (cr, colors, &params, &handle,
		                             x, y, width, height);
	}
	else if (DETAIL ("paned"))
	{
		WidgetParameters params;
		HandleParameters handle;

		excelsior_set_widget_parameters (widget, style, state_type, &params);
		handle.type = XCL_HANDLE_SPLITTER;
		handle.horizontal = (orientation == GTK_ORIENTATION_HORIZONTAL);

		STYLE_FUNCTION(draw_handle) (cr, colors, &params, &handle,
		                             x, y, width, height);
	}
	else
	{
		WidgetParameters params;
		HandleParameters handle;

		excelsior_set_widget_parameters (widget, style, state_type, &params);
		handle.type = XCL_HANDLE_TOOLBAR;
		handle.horizontal = (orientation == GTK_ORIENTATION_HORIZONTAL);

		STYLE_FUNCTION(draw_handle) (cr, colors, &params, &handle,
		                             x, y, width, height);
	}

	cairo_destroy (cr);
}

static void
excelsior_style_draw_box (DRAW_ARGS)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
	const excelsiorColors *colors;
	cairo_t *cr;

	cr     = ge_gdk_drawable_to_cairo (window, area);
	colors = &excelsior_style->colors;

	CHECK_ARGS
	SANITIZE_SIZE

	if (DETAIL ("menubar"))
	{
		WidgetParameters params;
		MenuBarParameters menubar;
		gboolean horizontal;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		menubar.style = excelsior_style->menubarstyle;

		horizontal = height < 2*width;
		/* This is not that great. Ideally we would have a nice vertical menubar. */
		if ((shadow_type != GTK_SHADOW_NONE) && horizontal)
			STYLE_FUNCTION(draw_menubar) (cr, colors, &params, &menubar,
			                              x, y, width, height);
	}
	else if (DETAIL ("button") && CHECK_HINT (GE_HINT_TREEVIEW_HEADER))
	{
		WidgetParameters params;
		ListViewHeaderParameters header;

		gint columns, column_index;
		gboolean resizable = TRUE;

		columns = 3;
		column_index = 1;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		params.corners = CR_CORNER_NONE;

		if (GE_IS_TREE_VIEW (widget->parent))
		{
			excelsior_treeview_get_header_index (GTK_TREE_VIEW(widget->parent),
			                                      widget, &column_index, &columns,
			                                      &resizable);
		}
		else if (GE_IS_CLIST (widget->parent))
		{
			excelsior_clist_get_header_index (GTK_CLIST(widget->parent),
			                                   widget, &column_index, &columns);
		}

		header.resizable = resizable;

		if (column_index == 0)
			header.order = params.ltr ? XCL_ORDER_FIRST : XCL_ORDER_LAST;
		else if (column_index == columns-1)
			header.order = params.ltr ? XCL_ORDER_LAST : XCL_ORDER_FIRST;
		else
			header.order = XCL_ORDER_MIDDLE;

		gtk_style_apply_default_background (style, window, FALSE, state_type, area, x, y, width, height);

		STYLE_FUNCTION(draw_list_view_header) (cr, colors, &params, &header,
		                                       x, y, width, height);
	}
	else if (DETAIL ("button") || DETAIL ("buttondefault"))
	{
		WidgetParameters params;
		ShadowParameters shadow = { CR_CORNER_ALL, XCL_SHADOW_NONE } ;
		excelsior_set_widget_parameters (widget, style, state_type, &params);
		params.active = shadow_type == GTK_SHADOW_IN;

		if (CHECK_HINT (GE_HINT_COMBOBOX_ENTRY))
		{
			if (params.ltr)
				params.corners = CR_CORNER_TOPRIGHT | CR_CORNER_BOTTOMRIGHT;
			else
				params.corners = CR_CORNER_TOPLEFT | CR_CORNER_BOTTOMLEFT;

			shadow.shadow = XCL_SHADOW_IN;

			if (params.ltr)
				x--;
			width++;
		}
		else
		{
			params.corners = CR_CORNER_ALL;
		}

		STYLE_FUNCTION(draw_button) (cr, &excelsior_style->colors, &params,
		                             x, y, width, height);

	}
	else if (DETAIL ("spinbutton_up") || DETAIL ("spinbutton_down"))
	{
		if (state_type == GTK_STATE_ACTIVE)
		{
			WidgetParameters params;
			excelsior_set_widget_parameters (widget, style, state_type, &params);

			if (style->xthickness == 3)
			{
				width++;
				if (params.ltr)
					x--;
			}

			if (DETAIL ("spinbutton_up"))
			{
				height+=2;
				if (params.ltr)
					params.corners = CR_CORNER_TOPRIGHT;
				else
					params.corners = CR_CORNER_TOPLEFT;
			}
			else
			{
				if (params.ltr)
					params.corners = CR_CORNER_BOTTOMRIGHT;
				else
					params.corners = CR_CORNER_BOTTOMLEFT;
			}

			STYLE_FUNCTION(draw_spinbutton_down) (cr, &excelsior_style->colors, &params, x, y, width, height);
		}
	}
	else if (DETAIL ("spinbutton"))
	{
		WidgetParameters params;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		if (params.ltr)
			params.corners = CR_CORNER_TOPRIGHT | CR_CORNER_BOTTOMRIGHT;
		else
			params.corners = CR_CORNER_TOPLEFT | CR_CORNER_BOTTOMLEFT;

		if (style->xthickness == 3)
		{
			if (params.ltr)
				x--;
			width++;
		}

		STYLE_FUNCTION(draw_spinbutton) (cr, &excelsior_style->colors, &params,
		                                 x, y, width, height);
	}
	else if (detail && g_str_has_prefix (detail, "trough") && CHECK_HINT (GE_HINT_SCALE))
	{
		WidgetParameters params;
		SliderParameters slider;

		excelsior_set_widget_parameters (widget, style, state_type, &params);
		params.corners    = CR_CORNER_NONE;

		slider.lower = DETAIL ("trough-lower");
		slider.fill_level = DETAIL ("trough-fill-level") || DETAIL ("trough-fill-level-full");

		if (CHECK_HINT (GE_HINT_HSCALE))
			slider.horizontal = TRUE;
		else if (CHECK_HINT (GE_HINT_VSCALE))
			slider.horizontal = FALSE;
		else /* Fallback based on the size... */
			slider.horizontal = width >= height;

		STYLE_FUNCTION(draw_scale_trough) (cr, &excelsior_style->colors,
		                                   &params, &slider,
		                                   x, y, width, height);
	}
	else if (DETAIL ("trough") && CHECK_HINT (GE_HINT_PROGRESSBAR))
	{
		WidgetParameters params;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		STYLE_FUNCTION(draw_progressbar_trough) (cr, colors, &params,
		                                         x, y, width, height);
	}
	else if (DETAIL ("trough") && CHECK_HINT (GE_HINT_SCROLLBAR))
	{
		WidgetParameters params;
		ScrollBarParameters scrollbar;

		excelsior_set_widget_parameters (widget, style, state_type, &params);
		params.corners = CR_CORNER_ALL;

		scrollbar.horizontal = TRUE;
		scrollbar.junction   = excelsior_scrollbar_get_junction (widget);


		if (CHECK_HINT (GE_HINT_HSCROLLBAR))
			scrollbar.horizontal = TRUE;
		else if (CHECK_HINT (GE_HINT_VSCROLLBAR))
			scrollbar.horizontal = FALSE;
		else /* Fallback based on the size  ... */
			scrollbar.horizontal = width >= height;

		/* What is this about? */
		if (scrollbar.horizontal)
		{
			x += 2;
			width -= 4;
		}
		else
		{
			y += 2;
			height -= 4;
		}

		STYLE_FUNCTION(draw_scrollbar_trough) (cr, colors, &params, &scrollbar,
		                                       x, y, width, height);
	}
	else if (DETAIL ("bar"))
	{
		WidgetParameters      params;
		ProgressBarParameters progressbar;
		gdouble               elapsed = 0.0;

#ifdef HAVE_ANIMATION
		if(excelsior_style->animation && XCL_IS_PROGRESS_BAR (widget))
		{
			gboolean activity_mode = GTK_PROGRESS (widget)->activity_mode;

			if (!activity_mode)
				excelsior_animation_progressbar_add ((gpointer)widget);
		}

		elapsed = excelsior_animation_elapsed (widget);
#endif

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		if (widget && GE_IS_PROGRESS_BAR (widget))
		{
			progressbar.orientation = gtk_progress_bar_get_orientation (GTK_PROGRESS_BAR (widget));
			progressbar.value = gtk_progress_bar_get_fraction(GTK_PROGRESS_BAR(widget));
			progressbar.pulsing = GTK_PROGRESS (widget)->activity_mode;
		}
		else
		{
			progressbar.orientation = XCL_ORIENTATION_LEFT_TO_RIGHT;
			progressbar.value = 0;
			progressbar.pulsing = FALSE;
		}

		if (!params.ltr)
		{
			if (progressbar.orientation == GTK_PROGRESS_LEFT_TO_RIGHT)
				progressbar.orientation = GTK_PROGRESS_RIGHT_TO_LEFT;
			else if (progressbar.orientation == GTK_PROGRESS_RIGHT_TO_LEFT)
				progressbar.orientation = GTK_PROGRESS_LEFT_TO_RIGHT;
		}

		/* Following is a hack to have a larger clip area, the one passed in
		 * does not allow for the shadow. */
		if (area)
		{
			GdkRectangle tmp = *area;
			if (!progressbar.pulsing)
			{
				switch (progressbar.orientation)
				{
					case GTK_PROGRESS_RIGHT_TO_LEFT:
						tmp.x -= 1;
					case GTK_PROGRESS_LEFT_TO_RIGHT:
						tmp.width += 1;
						break;
					case GTK_PROGRESS_BOTTOM_TO_TOP:
						tmp.y -= 1;
					case GTK_PROGRESS_TOP_TO_BOTTOM:
						tmp.height += 1;
						break;
				}
			}
			else
			{
				if (progressbar.orientation == GTK_PROGRESS_RIGHT_TO_LEFT ||
				    progressbar.orientation == GTK_PROGRESS_LEFT_TO_RIGHT)
				{
					tmp.x -= 1;
					tmp.width += 2;
				}
				else
				{
					tmp.y -= 1;
					tmp.height += 2;
				}
			}

			cairo_reset_clip (cr);
			gdk_cairo_rectangle (cr, &tmp);
			cairo_clip (cr);
		}

		progressbar.progbar_to_left = excelsior_style->progbar_to_left;
		progressbar.progbar_style = excelsior_style->progbar_style;

		if (progressbar.progbar_to_left == 0)
		{
			STYLE_FUNCTION(draw_progressbar_fill) (cr, colors, &params, &progressbar,
		                                       x, y, width, height,
		                                       10 - (int)(elapsed * 10.0) % 10);
		}
		else if (progressbar.progbar_to_left == 1)
		{
			STYLE_FUNCTION(draw_progressbar_fill) (cr, colors, &params, &progressbar,
		                                       x, y, width, height,
		                                       10 + (int)(elapsed * 10.0) % 10);
		}

	}
	else if (DETAIL ("optionmenu"))
	{
		WidgetParameters params;
		OptionMenuParameters optionmenu;

		GtkRequisition indicator_size;
		GtkBorder indicator_spacing;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		ge_option_menu_get_props (widget, &indicator_size, &indicator_spacing);

		if (ge_widget_is_ltr (widget))
			optionmenu.linepos = width - (indicator_size.width + indicator_spacing.left + indicator_spacing.right) - 1;
		else
			optionmenu.linepos = (indicator_size.width + indicator_spacing.left + indicator_spacing.right) + 1;

		STYLE_FUNCTION(draw_optionmenu) (cr, colors, &params, &optionmenu,
		                                 x, y, width, height);
	}
	else if (DETAIL ("menuitem"))
	{
		WidgetParameters params;
		excelsior_set_widget_parameters (widget, style, state_type, &params);

		if (CHECK_HINT (GE_HINT_MENUBAR))
		{
			params.corners = CR_CORNER_TOPLEFT | CR_CORNER_TOPRIGHT;
			height += 1;
			STYLE_FUNCTION(draw_menubaritem) (cr, colors, &params, x, y, width, height);
		}
		else
		{
			params.corners = CR_CORNER_ALL;
			STYLE_FUNCTION(draw_menuitem) (cr, colors, &params, x, y, width, height);
		}
	}
	else if (DETAIL ("hscrollbar") || DETAIL ("vscrollbar")) /* This can't be "stepper" for scrollbars ... */
	{
		WidgetParameters    params;
		ScrollBarParameters scrollbar;
		ScrollBarStepperParameters stepper;
		GdkRectangle this_rectangle;

		this_rectangle.x = x;
		this_rectangle.y = y;
		this_rectangle.width  = width;
		this_rectangle.height = height;

		excelsior_set_widget_parameters (widget, style, state_type, &params);
		params.corners = CR_CORNER_NONE;

		scrollbar.style = excelsior_style->scrollbarstyle;
		scrollbar.horizontal = TRUE;
		scrollbar.junction   = excelsior_scrollbar_get_junction (widget);

		scrollbar.horizontal = DETAIL ("hscrollbar");

		stepper.stepper = excelsior_scrollbar_get_stepper (widget, &this_rectangle);

		STYLE_FUNCTION(draw_scrollbar_stepper) (cr, colors, &params, &scrollbar, &stepper,
		                                        x, y, width, height);
	}
	else if (DETAIL ("toolbar") || DETAIL ("handlebox_bin") || DETAIL ("dockitem_bin"))
	{
		WidgetParameters  params;
		ToolbarParameters toolbar;
		gboolean horizontal;

		excelsior_set_widget_parameters (widget, style, state_type, &params);
		excelsior_set_toolbar_parameters (&toolbar, widget, window, x, y);

		toolbar.style = excelsior_style->toolbarstyle;

		if ((DETAIL ("handlebox_bin") || DETAIL ("dockitem_bin")) && GE_IS_BIN (widget))
		{
			GtkWidget* child = gtk_bin_get_child ((GtkBin*) widget);
			/* This is to draw the correct shadow on the handlebox.
			 * We need to draw it here, as otherwise the handle will not get the
			 * background. */
			if (GE_IS_TOOLBAR (child))
				gtk_widget_style_get (child, "shadow-type", &shadow_type, NULL);
		}

		horizontal = height < 2*width;
		/* This is not that great. Ideally we would have a nice vertical toolbar. */
		if ((shadow_type != GTK_SHADOW_NONE) && horizontal)
			STYLE_FUNCTION(draw_toolbar) (cr, colors, &params, &toolbar, x, y, width, height);
	}
	else if (DETAIL ("menu"))
	{
		WidgetParameters params;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		STYLE_FUNCTION(draw_menu_frame) (cr, colors, &params, x, y, width, height);
	}
	else if (DETAIL ("hseparator") || DETAIL ("vseparator"))
	{
		WidgetParameters params;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		gchar *new_detail = (gchar*) detail;
		/* Draw a normal separator, we just use this because it gives more control
		 * over sizing (currently). */

		/* This isn't nice ... but it seems like the best cleanest way to me right now.
		 * It will get slightly nicer in the future hopefully. */
		if (GE_IS_MENU_ITEM (widget))
			new_detail = "menuitem";

		if (DETAIL ("hseparator"))
		{
			gtk_paint_hline (style, window, state_type, area, widget, new_detail,
			                 x, x + width - 1, y + height/2);
		}
		else
			gtk_paint_vline (style, window, state_type, area, widget, new_detail,
		                 y, y + height - 1, x + width/2);

	}
	else
	{
		excelsior_parent_class->draw_box (style, window, state_type, shadow_type, area,
		                                   widget, detail, x, y, width, height);
	}

	cairo_destroy (cr);
}

static void
excelsior_style_draw_slider (DRAW_ARGS, GtkOrientation orientation)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
	const excelsiorColors *colors;
	cairo_t *cr;

	cr     = ge_gdk_drawable_to_cairo (window, area);
	colors = &excelsior_style->colors;

	CHECK_ARGS
	SANITIZE_SIZE

	if (DETAIL ("hscale") || DETAIL ("vscale"))
	{
		WidgetParameters params;
		SliderParameters slider;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		slider.horizontal = (orientation == GTK_ORIENTATION_HORIZONTAL);
		slider.lower = FALSE;
		slider.fill_level = FALSE;

		STYLE_FUNCTION(draw_slider_button) (cr, &excelsior_style->colors,
		                                    &params, &slider,
		                                    x, y, width, height);
	}
	else if (DETAIL ("slider"))
	{
		WidgetParameters    params;
		ScrollBarParameters scrollbar;

		excelsior_set_widget_parameters (widget, style, state_type, &params);
		params.corners = CR_CORNER_NONE;

		scrollbar.horizontal = (orientation == GTK_ORIENTATION_HORIZONTAL);
		scrollbar.junction   = excelsior_scrollbar_get_junction (widget);

		scrollbar.style = excelsior_style->scrollbarstyle;


		STYLE_FUNCTION(draw_scrollbar_slider) (cr, colors, &params, &scrollbar,
		                                       x, y, width, height);
	}
	else
	{
		excelsior_parent_class->draw_slider (style, window, state_type, shadow_type, area,
		                                      widget, detail, x, y, width, height, orientation);
	}

	cairo_destroy (cr);
}

static void
excelsior_style_draw_option (DRAW_ARGS)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
	const excelsiorColors *colors;
	WidgetParameters params;
	CheckboxParameters checkbox;
	cairo_t *cr;

	CHECK_ARGS
	SANITIZE_SIZE

	cr = ge_gdk_drawable_to_cairo (window, area);
	colors = &excelsior_style->colors;

	checkbox.shadow_type = shadow_type;
	checkbox.in_menu = (widget && GTK_IS_MENU(widget->parent));

	excelsior_set_widget_parameters (widget, style, state_type, &params);

	STYLE_FUNCTION(draw_radiobutton) (cr, colors, &params, &checkbox, x, y, width, height);

	cairo_destroy (cr);
}

static void
excelsior_style_draw_check (DRAW_ARGS)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
	WidgetParameters params;
	CheckboxParameters checkbox;
	cairo_t *cr;

	CHECK_ARGS
	SANITIZE_SIZE

	cr = ge_gdk_drawable_to_cairo (window, area);

	excelsior_set_widget_parameters (widget, style, state_type, &params);

	params.corners = CR_CORNER_ALL;

	checkbox.shadow_type = shadow_type;
	checkbox.in_cell = DETAIL("cellcheck");

	checkbox.in_menu = (widget && widget->parent && GTK_IS_MENU(widget->parent));

	STYLE_FUNCTION(draw_checkbox) (cr, &excelsior_style->colors, &params, &checkbox,
	                               x, y, width, height);

	cairo_destroy (cr);
}

static void
excelsior_style_draw_vline (GtkStyle               *style,
                             GdkWindow              *window,
                             GtkStateType            state_type,
                             GdkRectangle           *area,
                             GtkWidget              *widget,
                             const gchar            *detail,
                             gint                    y1,
                             gint                    y2,
                             gint                    x)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
	const excelsiorColors *colors;
	SeparatorParameters separator = { FALSE };
	cairo_t *cr;

	CHECK_ARGS

	colors = &excelsior_style->colors;

	cr = ge_gdk_drawable_to_cairo (window, area);

	if (GE_IS_COMBO_BOX (widget->parent->parent->parent))
	{
	}
	else
	{
	STYLE_FUNCTION(draw_separator) (cr, colors, NULL, &separator,
	                                x, y1, 2, y2-y1+1);
	}

	cairo_destroy (cr);
}

static void
excelsior_style_draw_hline (GtkStyle               *style,
                             GdkWindow              *window,
                             GtkStateType            state_type,
                             GdkRectangle           *area,
                             GtkWidget              *widget,
                             const gchar            *detail,
                             gint                    x1,
                             gint                    x2,
                             gint                    y)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
	const excelsiorColors *colors;
	cairo_t *cr;
	SeparatorParameters separator;

	CHECK_ARGS

	colors = &excelsior_style->colors;

	cr = ge_gdk_drawable_to_cairo (window, area);

	separator.horizontal = TRUE;

	if (!DETAIL ("menuitem"))
		STYLE_FUNCTION(draw_separator) (cr, colors, NULL, &separator,
		                                x1, y, x2-x1+1, 2);
	else
		STYLE_FUNCTION(draw_menu_item_separator) (cr, colors, NULL, &separator,
		                                          x1, y, x2-x1+1, 2);

	cairo_destroy (cr);
}

static void
excelsior_style_draw_shadow_gap (DRAW_ARGS,
                                  GtkPositionType gap_side,
                                  gint            gap_x,
                                  gint            gap_width)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
	const excelsiorColors *colors;
	cairo_t *cr;

	CHECK_ARGS
	SANITIZE_SIZE

	cr     = ge_gdk_drawable_to_cairo (window, area);
	colors = &excelsior_style->colors;

	if (DETAIL ("frame"))
	{
		WidgetParameters params;
		FrameParameters  frame;

		frame.shadow    = shadow_type;
		frame.gap_side  = gap_side;
		frame.gap_x     = gap_x;
		frame.gap_width = gap_width;
		frame.border    = &colors->shade[5];

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		params.corners = CR_CORNER_ALL;

		STYLE_FUNCTION(draw_frame) (cr, colors, &params, &frame,
		                            x, y, width, height);
	}
	else
	{
		excelsior_parent_class->draw_shadow_gap (style, window, state_type, shadow_type, area,
		                                          widget, detail, x, y, width, height,
		                                          gap_side, gap_x, gap_width);
	}

	cairo_destroy (cr);
}

static void
excelsior_style_draw_resize_grip (GtkStyle       *style,
                                   GdkWindow      *window,
                                   GtkStateType    state_type,
                                   GdkRectangle   *area,
                                   GtkWidget      *widget,
                                   const gchar    *detail,
                                   GdkWindowEdge   edge,
                                   gint            x,
                                   gint            y,
                                   gint            width,
                                   gint            height)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
	excelsiorColors *colors = &excelsior_style->colors;
	cairo_t *cr;
	WidgetParameters params;
	ResizeGripParameters grip;

	CHECK_ARGS
	SANITIZE_SIZE

	grip.edge = (excelsiorWindowEdge)edge;

	g_return_if_fail (window != NULL);

	cr = ge_gdk_drawable_to_cairo (window, area);

	excelsior_set_widget_parameters (widget, style, state_type, &params);

	STYLE_FUNCTION(draw_resize_grip) (cr, colors, &params, &grip,
	                                  x, y, width, height);

	cairo_destroy (cr);
}

static void
excelsior_style_draw_tab (DRAW_ARGS)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
	excelsiorColors *colors = &excelsior_style->colors;
	WidgetParameters params;
	ArrowParameters  arrow;
	cairo_t *cr;

	CHECK_ARGS
	SANITIZE_SIZE

	cr = ge_gdk_drawable_to_cairo (window, area);

	excelsior_set_widget_parameters (widget, style, state_type, &params);
	arrow.type      = XCL_ARROW_COMBO;
	arrow.direction = XCL_DIRECTION_DOWN;

	STYLE_FUNCTION(draw_arrow) (cr, colors, &params, &arrow, x, y, width, height);


	cairo_destroy (cr);
}

static void
excelsior_style_draw_expander (GtkStyle * style,
			    GdkWindow * window,
			    GtkStateType state_type,
			    GdkRectangle * area,
			    GtkWidget * widget,
			    const gchar * detail,
			    gint width,
			    gint height,
			    GtkExpanderStyle expander_style)
{
	cairo_t *cr;
	gint expander_size;
	cr = gdk_cairo_create (window);
	expander_size = 8;

	CHECK_ARGS
	SANITIZE_SIZE

	cairo_translate (cr, width, height);

	switch (expander_style) {
		case GTK_EXPANDER_COLLAPSED:
			cairo_move_to (cr, 0, -4);
			cairo_line_to (cr, expander_size, expander_size/4-2);
			cairo_line_to (cr, 0, expander_size/2);
			cairo_close_path (cr);

			gdk_cairo_set_source_color (cr, &style->fg[GTK_STATE_NORMAL]);
		break;
		case GTK_EXPANDER_SEMI_COLLAPSED:
			cairo_move_to (cr, 0, -4);
			cairo_line_to (cr, expander_size, expander_size/4-2);
			cairo_line_to (cr, 0, expander_size/2);
			cairo_close_path (cr);

			gdk_cairo_set_source_color (cr, &style->fg[GTK_STATE_NORMAL]);
		break;
		case GTK_EXPANDER_SEMI_EXPANDED:
			cairo_move_to (cr, 0, -4);
			cairo_line_to (cr, expander_size, -4);
			cairo_line_to (cr, expander_size/2, -4+expander_size);
			cairo_close_path (cr);

			gdk_cairo_set_source_color (cr, &style->bg[GTK_STATE_SELECTED]);
		break;
		case GTK_EXPANDER_EXPANDED:
			cairo_move_to (cr, 0, -4);
			cairo_line_to (cr, expander_size, -4);
			cairo_line_to (cr, expander_size/2, -4+expander_size);
			cairo_close_path (cr);

			gdk_cairo_set_source_color (cr, &style->bg[GTK_STATE_SELECTED]);
		break;
		default:
			g_assert_not_reached ();
	}

	cairo_set_line_width (cr, 1.0);

	cairo_fill_preserve (cr);

	cairo_destroy (cr);
}

static void
excelsior_style_draw_arrow (GtkStyle  *style,
                             GdkWindow     *window,
                             GtkStateType   state_type,
                             GtkShadowType  shadow,
                             GdkRectangle  *area,
                             GtkWidget     *widget,
                             const gchar   *detail,
                             GtkArrowType   arrow_type,
                             gboolean       fill,
                             gint           x,
                             gint           y,
                             gint           width,
                             gint           height)
{
	excelsiorStyle  *excelsior_style = EXCELSIOR_STYLE (style);
	excelsiorColors *colors = &excelsior_style->colors;
	WidgetParameters params;
	ArrowParameters  arrow;
	cairo_t *cr = ge_gdk_drawable_to_cairo (window, area);

	CHECK_ARGS
	SANITIZE_SIZE

	if (arrow_type == GTK_ARROW_NONE)
	{
		cairo_destroy (cr);
		return;
	}

	excelsior_set_widget_parameters (widget, style, state_type, &params);
	arrow.type = XCL_ARROW_NORMAL;
	arrow.direction = (excelsiorDirection)arrow_type;

	if (widget && widget->parent && GE_IS_COMBO (widget->parent->parent))
	{
		x += 1;
		width -= 2;
	}

	if (ge_is_combo_box (widget, FALSE) && !ge_is_combo_box_entry (widget))
	{
		arrow.type = XCL_ARROW_COMBO;
		STYLE_FUNCTION(draw_arrow) (cr, colors, &params, &arrow, x+1, y, width, height);
	}
	else if (GE_IS_MENU(widget->parent) && GTK_MINOR_VERSION >= 13)
	{
		STYLE_FUNCTION(draw_arrow) (cr, colors, &params, &arrow, x, y, width-2, height-2);
	}
	else
	{
		STYLE_FUNCTION(draw_arrow) (cr, colors, &params, &arrow, x, y, width, height);
	}

	cairo_destroy (cr);
}

static void
excelsior_style_init_from_rc (GtkStyle * style,
                               GtkRcStyle * rc_style)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);

	excelsior_parent_class->init_from_rc (style, rc_style);

	g_assert ((EXCELSIOR_RC_STYLE (rc_style)->style >= 0) && (EXCELSIOR_RC_STYLE (rc_style)->style < XCL_NUM_STYLES));
	excelsior_style->style               = EXCELSIOR_RC_STYLE (rc_style)->style;

	excelsior_style->menubarstyle        = EXCELSIOR_RC_STYLE (rc_style)->menubarstyle;
	excelsior_style->toolbarstyle        = EXCELSIOR_RC_STYLE (rc_style)->toolbarstyle;
	excelsior_style->scrollbarstyle      = EXCELSIOR_RC_STYLE (rc_style)->scrollbarstyle;
	excelsior_style->progbar_style       = EXCELSIOR_RC_STYLE (rc_style)->progbar_style;
	excelsior_style->progbar_to_left     = EXCELSIOR_RC_STYLE (rc_style)->progbar_to_left;
	excelsior_style->random_effect       = EXCELSIOR_RC_STYLE (rc_style)->random_effect;
	excelsior_style->random_opacity      = CLAMP (EXCELSIOR_RC_STYLE (rc_style)->random_opacity, 0.0, 1.0);
	excelsior_style->animation           = EXCELSIOR_RC_STYLE (rc_style)->animation;
	excelsior_style->mw_gradient         = EXCELSIOR_RC_STYLE (rc_style)->mw_gradient;
	excelsior_style->aqua_tabs           = EXCELSIOR_RC_STYLE (rc_style)->aqua_tabs;
	excelsior_style->pellucid_scales     = EXCELSIOR_RC_STYLE (rc_style)->pellucid_scales;
	excelsior_style->scale_trough_size   = EXCELSIOR_RC_STYLE (rc_style)->scale_trough_size;
	excelsior_style->tooltip_style       = EXCELSIOR_RC_STYLE (rc_style)->tooltip_style;
	excelsior_style->radius              = CLAMP (EXCELSIOR_RC_STYLE (rc_style)->radius, 0.0, 10.0);
	excelsior_style->grip_opacity        = CLAMP (EXCELSIOR_RC_STYLE (rc_style)->grip_opacity, 0.0, 1.0);
	excelsior_style->grip_style          = EXCELSIOR_RC_STYLE (rc_style)->grip_style;
	excelsior_style->trough_visible      = EXCELSIOR_RC_STYLE (rc_style)->trough_visible;
}

static void
excelsior_style_realize (GtkStyle * style)
{
	excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
	double shades[] = {1.15, 0.95, 0.896, 0.82, 0.7, 0.665, 0.475, 0.45, 0.4};
	CairoColor spot_color;
	CairoColor bg_normal;
	double contrast;
	int i;

	excelsior_parent_class->realize (style);

	contrast = EXCELSIOR_RC_STYLE (style->rc_style)->contrast;

	/* Lighter to darker */
	ge_gdk_color_to_cairo (&style->bg[GTK_STATE_NORMAL], &bg_normal);

	for (i = 0; i < 9; i++)
	{
		ge_shade_color (&bg_normal, (shades[i] < 1.0) ?
		                (shades[i]/contrast) : (shades[i]*contrast),
		                &excelsior_style->colors.shade[i]);
	}

	ge_gdk_color_to_cairo (&style->bg[GTK_STATE_SELECTED], &spot_color);

	ge_shade_color (&spot_color, 1.25, &excelsior_style->colors.spot[0]);
	ge_shade_color (&spot_color, 1.05, &excelsior_style->colors.spot[1]);
	ge_shade_color (&spot_color, 0.65, &excelsior_style->colors.spot[2]);

	for (i=0; i<5; i++)
	{
		ge_gdk_color_to_cairo (&style->fg[i], &excelsior_style->colors.fg[i]);
		ge_gdk_color_to_cairo (&style->bg[i], &excelsior_style->colors.bg[i]);
		ge_gdk_color_to_cairo (&style->base[i], &excelsior_style->colors.base[i]);
		ge_gdk_color_to_cairo (&style->text[i], &excelsior_style->colors.text[i]);
	}
}

static void
excelsior_style_copy (GtkStyle * style, GtkStyle * src)
{
	excelsiorStyle * xcl_style = EXCELSIOR_STYLE (style);
	excelsiorStyle * xcl_src = EXCELSIOR_STYLE (src);

	xcl_style->colors              = xcl_src->colors;
	xcl_style->menubarstyle        = xcl_src->menubarstyle;
	xcl_style->toolbarstyle        = xcl_src->toolbarstyle;
	xcl_style->animation           = xcl_src->animation;
	xcl_style->scrollbarstyle      = xcl_src->scrollbarstyle;
	xcl_style->progbar_style       = xcl_src->progbar_style;
	xcl_style->progbar_to_left     = xcl_src->progbar_to_left;
	xcl_style->random_effect       = xcl_src->random_effect;
	xcl_style->radius              = xcl_src->radius;
	xcl_style->random_effect       = xcl_src->random_effect;
	xcl_style->random_opacity      = xcl_src->random_opacity;
	xcl_style->mw_gradient         = xcl_src->mw_gradient;
	xcl_style->aqua_tabs           = xcl_src->aqua_tabs;
	xcl_style->style               = xcl_src->style;
	xcl_style->pellucid_scales     = xcl_src->pellucid_scales;
	xcl_style->scale_trough_size   = xcl_src->scale_trough_size;
	xcl_style->tooltip_style       = xcl_src->tooltip_style;
	xcl_style->grip_opacity        = xcl_src->grip_opacity;
	xcl_style->grip_style          = xcl_src->grip_style;
	xcl_style->trough_visible      = xcl_src->trough_visible;

	excelsior_parent_class->copy (style, src);
}

static void
excelsior_style_unrealize (GtkStyle * style)
{
	excelsior_parent_class->unrealize (style);
}

static GdkPixbuf *
set_transparency (const GdkPixbuf *pixbuf, gdouble alpha_percent)
{
	GdkPixbuf *target;
	guchar *data, *current;
	guint x, y, rowstride, height, width;

	g_return_val_if_fail (pixbuf != NULL, NULL);
	g_return_val_if_fail (GDK_IS_PIXBUF (pixbuf), NULL);

	/* Returns a copy of pixbuf with it's non-completely-transparent pixels to
	   have an alpha level "alpha_percent" of their original value. */

	target = gdk_pixbuf_add_alpha (pixbuf, FALSE, 0, 0, 0);

	if (alpha_percent == 1.0)
		return target;
	width = gdk_pixbuf_get_width (target);
	height = gdk_pixbuf_get_height (target);
	rowstride = gdk_pixbuf_get_rowstride (target);
	data = gdk_pixbuf_get_pixels (target);

	for (y = 0; y < height; y++)
	{
		for (x = 0; x < width; x++)
		{
			/* The "4" is the number of chars per pixel, in this case, RGBA,
			   the 3 means "skip to the alpha" */
			current = data + (y * rowstride) + (x * 4) + 3;
			*(current) = (guchar) (*(current) * alpha_percent);
		}
	}

	return target;
}

static GdkPixbuf*
scale_or_ref (GdkPixbuf *src,
              int width,
              int height)
{
	if (width == gdk_pixbuf_get_width (src) &&
	    height == gdk_pixbuf_get_height (src))
	{
		return g_object_ref (src);
	}
	else
	{
		return gdk_pixbuf_scale_simple (src,
		                                width, height,
		                                GDK_INTERP_BILINEAR);
	}
}

static void
excelsior_style_draw_layout (GtkStyle * style,
                              GdkWindow * window,
                              GtkStateType state_type,
                              gboolean use_text,
                              GdkRectangle * area,
                              GtkWidget * widget,
                              const gchar * detail, gint x, gint y, PangoLayout * layout)
{
	GdkGC *gc;

	g_return_if_fail (GTK_IS_STYLE (style));
	g_return_if_fail (window != NULL);

	gc = use_text ? style->text_gc[state_type] : style->fg_gc[state_type];

	if (area)
		gdk_gc_set_clip_rectangle (gc, area);

	if (state_type == GTK_STATE_INSENSITIVE)
	{
		excelsiorStyle *excelsior_style = EXCELSIOR_STYLE (style);
		excelsiorColors *colors = &excelsior_style->colors;

		WidgetParameters params;
		GdkColor etched;
		CairoColor temp;

		excelsior_set_widget_parameters (widget, style, state_type, &params);

		ge_shade_color (&colors->bg[widget->state], 1.2, &temp);

		etched.red = (int) (temp.r * 65535);
		etched.green = (int) (temp.g * 65535);
		etched.blue = (int) (temp.b * 65535);

		gdk_draw_layout_with_colors (window, gc, x + 1, y + 1, layout, &etched, NULL);
		gdk_draw_layout (window, gc, x, y, layout);
	}
	else
		gdk_draw_layout (window, gc, x, y, layout);

	if (area)
		gdk_gc_set_clip_rectangle (gc, NULL);
}

static GdkPixbuf *
excelsior_style_draw_render_icon (GtkStyle            *style,
                                   const GtkIconSource *source,
                                   GtkTextDirection     direction,
                                   GtkStateType         state,
                                   GtkIconSize          size,
                                   GtkWidget           *widget,
                                   const char          *detail)
{
	int width = 1;
	int height = 1;
	GdkPixbuf *scaled;
	GdkPixbuf *stated;
	GdkPixbuf *base_pixbuf;
	GdkScreen *screen;
	GtkSettings *settings;

	/* Oddly, style can be NULL in this function, because
	 * GtkIconSet can be used without a style and if so
	 * it uses this function.
	 */

	base_pixbuf = gtk_icon_source_get_pixbuf (source);

	g_return_val_if_fail (base_pixbuf != NULL, NULL);

	if (widget && gtk_widget_has_screen (widget))
	{
		screen = gtk_widget_get_screen (widget);
		settings = gtk_settings_get_for_screen (screen);
	}
	else if (style->colormap)
	{
		screen = gdk_colormap_get_screen (style->colormap);
		settings = gtk_settings_get_for_screen (screen);
	}
	else
	{
		settings = gtk_settings_get_default ();
		GTK_NOTE (MULTIHEAD,
			  g_warning ("Using the default screen for gtk_default_render_icon()"));
	}

	if (size != (GtkIconSize) -1 && !gtk_icon_size_lookup_for_settings (settings, size, &width, &height))
	{
		g_warning (G_STRLOC ": invalid icon size '%d'", size);
		return NULL;
	}

	/* If the size was wildcarded, and we're allowed to scale, then scale; otherwise,
	 * leave it alone.
	 */
	if (size != (GtkIconSize)-1 && gtk_icon_source_get_size_wildcarded (source))
		scaled = scale_or_ref (base_pixbuf, width, height);
	else
		scaled = g_object_ref (base_pixbuf);

	/* If the state was wildcarded, then generate a state. */
	if (gtk_icon_source_get_state_wildcarded (source))
	{
		if (state == GTK_STATE_INSENSITIVE)
		{
			stated = set_transparency (scaled, 0.3);
			gdk_pixbuf_saturate_and_pixelate (stated, stated, 0.1, FALSE);

			g_object_unref (scaled);
		}
		else if (state == GTK_STATE_PRELIGHT)
		{
			stated = gdk_pixbuf_copy (scaled);

			gdk_pixbuf_saturate_and_pixelate (scaled, stated, 1.2, FALSE);

			g_object_unref (scaled);
		}
		else
		{
			stated = scaled;
		}
	}
	else
		stated = scaled;

	return stated;
}

static void
excelsior_style_init (excelsiorStyle * style)
{
}

static void
excelsior_style_class_init (excelsiorStyleClass * klass)
{
	GtkStyleClass *style_class = GTK_STYLE_CLASS (klass);

	excelsior_style_class = EXCELSIOR_STYLE_CLASS (klass);
	excelsior_parent_class = g_type_class_peek_parent (klass);

	style_class->copy             = excelsior_style_copy;
	style_class->realize          = excelsior_style_realize;
	style_class->unrealize        = excelsior_style_unrealize;
	style_class->init_from_rc     = excelsior_style_init_from_rc;
	style_class->draw_handle      = excelsior_style_draw_handle;
	style_class->draw_slider      = excelsior_style_draw_slider;
	style_class->draw_shadow_gap  = excelsior_style_draw_shadow_gap;
	style_class->draw_box         = excelsior_style_draw_box;
	style_class->draw_shadow      = excelsior_style_draw_shadow;
	style_class->draw_box_gap     = excelsior_style_draw_box_gap;
	style_class->draw_extension   = excelsior_style_draw_extension;
	style_class->draw_option      = excelsior_style_draw_option;
	style_class->draw_check       = excelsior_style_draw_check;
	style_class->draw_flat_box    = excelsior_style_draw_flat_box;
	style_class->draw_vline       = excelsior_style_draw_vline;
	style_class->draw_hline       = excelsior_style_draw_hline;
	style_class->draw_resize_grip = excelsior_style_draw_resize_grip;
	style_class->draw_tab         = excelsior_style_draw_tab;
	style_class->draw_expander    = excelsior_style_draw_expander;
	style_class->draw_arrow       = excelsior_style_draw_arrow;
	style_class->draw_layout      = excelsior_style_draw_layout;
	style_class->render_icon      = excelsior_style_draw_render_icon;

	excelsior_register_style_inverted (&excelsior_style_class->style_functions[XCL_STYLE_INVERTED]);

	klass->style_functions[XCL_STYLE_BUBBLEGUM] = klass->style_functions[XCL_STYLE_INVERTED];
	excelsior_register_style_bubblegum (&excelsior_style_class->style_functions[XCL_STYLE_BUBBLEGUM]);

	klass->style_functions[XCL_STYLE_AQUA] = klass->style_functions[XCL_STYLE_INVERTED];
	excelsior_register_style_aqua (&excelsior_style_class->style_functions[XCL_STYLE_AQUA]);

	klass->style_functions[XCL_STYLE_CLARIUS] = klass->style_functions[XCL_STYLE_INVERTED];
	excelsior_register_style_clarius (&excelsior_style_class->style_functions[XCL_STYLE_CLARIUS]);

	klass->style_functions[XCL_STYLE_FLAT] = klass->style_functions[XCL_STYLE_INVERTED];
	excelsior_register_style_flat (&excelsior_style_class->style_functions[XCL_STYLE_FLAT]);

}

GType excelsior_type_style = 0;

void
excelsior_style_register_type (GTypeModule * module)
{
	static const GTypeInfo object_info =
	{
		sizeof (excelsiorStyleClass),
		(GBaseInitFunc) NULL,
		(GBaseFinalizeFunc) NULL,
		(GClassInitFunc) excelsior_style_class_init,
		NULL,         /* class_finalize */
		NULL,         /* class_data */
		sizeof (excelsiorStyle),
		0,            /* n_preallocs */
		(GInstanceInitFunc) excelsior_style_init,
		NULL
	};

	excelsior_type_style = g_type_module_register_type (module,
	                                                     GTK_TYPE_STYLE,
	                                                     "excelsiorStyle",
	                                                     &object_info, 0);
}
