/* Excelsior theme engine
 * Copyright (C) 2005 Richard Stellingwerff.
 * Copyright (C) 2007 Benjamin Berg
 * Copyright (C) 2007 Andrea Cimitan
 * Copyright (C) 2008 Christopher Bratusek
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Written by Owen Taylor <otaylor@redhat.com>
 * and by Alexander Larsson <alexl@redhat.com>
 * Modified by Richard Stellingwerff <remenic@gmail.com>
 */

#include <gtk/gtk.h>

#ifndef EXCELSIOR_STYLE_H
#define EXCELSIOR_STYLE_H

#include "animation.h"
#include "excelsior_types.h"

typedef struct _excelsiorStyle excelsiorStyle;
typedef struct _excelsiorStyleClass excelsiorStyleClass;

GE_INTERNAL extern GType excelsior_type_style;

#define EXCELSIOR_TYPE_STYLE              excelsior_type_style
#define EXCELSIOR_STYLE(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), EXCELSIOR_TYPE_STYLE, excelsiorStyle))
#define EXCELSIOR_STYLE_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), EXCELSIOR_TYPE_STYLE, excelsiorStyleClass))
#define EXCELSIOR_IS_STYLE(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), EXCELSIOR_TYPE_STYLE))
#define EXCELSIOR_IS_STYLE_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), EXCELSIOR_TYPE_STYLE))
#define EXCELSIOR_STYLE_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), EXCELSIOR_TYPE_STYLE, excelsiorStyleClass))

struct _excelsiorStyle
{
	GtkStyle parent_instance;

	excelsiorColors colors;

	excelsiorStyles style;

	guint8   menubarstyle;
	guint8   toolbarstyle;
	guint8   scrollbarstyle;
	guint8   progbar_style;
	guint8   progbar_to_left;
	guint8   random_effect;
	gboolean animation;
	gfloat   radius;
	gfloat   random_opacity;
	guint8   mw_gradient;
	guint8   aqua_tabs;
	guint8   pellucid_scales;
	guint8   scale_trough_size;
	guint8   tooltip_style;
	gfloat   grip_opacity;
	guint8   grip_style;
	guint8   trough_visible;
};

struct _excelsiorStyleClass
{
	GtkStyleClass parent_class;

	excelsiorStyleFunctions style_functions[XCL_NUM_STYLES];
};

GE_INTERNAL void excelsior_style_register_type (GTypeModule *module);

#endif /* EXCELSIOR_STYLE_H */
